from remotemanager import Dataset


class SanzuWrapper:
    """
    Decorator class to allow you to

    Wrap a function with `@SanzuFunction` to store:

    >>> from remotemanager import URL
    >>> url = URL(...)
    >>> @SanzuFunction(url=url)
    >>> def func(val):
    >>>     return val
    >>> func(val=3) # creates a Dataset and runs this function on `url`

    Call this function will transparently create a Dataset, execute
    the function remotely (synchronously), and return the result.

    The function should be called with explicitly named keyword arguments,
    but logic exists to attempt to match args to kwargs
    """

    def __init__(self, function, *args, **kwargs):
        self._ds = Dataset(function=function, *args, **kwargs)

    def __call__(self, *args, **kwargs):
        # to enable an *args call, iterate over the signature, attempting to inject
        # at each kwarg
        for i, arg in enumerate(args):
            # strip any type hinting. e.g. `x: int = 10` should become `x`
            argname = self.dataset.function.args[i].split()[0].strip(":")

            if argname in kwargs:
                raise ValueError(f"Got multiple values for arg {argname}")

            kwargs[argname] = arg

        runner = self._ds.append_run(kwargs, return_runner=True)
        runner.run()

        self._ds.wait()
        self._ds.fetch_results()

        return runner.result

    @property
    def dataset(self) -> Dataset:
        """
        Return the current Dataset used for this function

        Returns:
            (Dataset): associated Dataset
        """
        return self._ds


def SanzuFunction(*args, **kwargs):
    """
    Actual decorator wrapper for SanzuFunction

    In order to make a decorator callable, Python seems to require an actual function
    call that returns the class.

    Args:
        *args:
            args to pass through to the Dataset
        **kwargs:
            kwargs to pass through to the Dataset

    Returns:
        decorator
    """
    if len(args) > 0 and hasattr(args[0], "__call__"):
        # calling the decorator with no args places the function in the first arg
        return SanzuWrapper(args[0], **kwargs)

    # Otherwise, capture the function via standard decorator
    def decorate(function):
        return SanzuWrapper(function, *args, **kwargs)

    return decorate



import hashlib

from remotemanager.logging.utils import format_iterable


def generate_uuid(string: str) -> str:
    if not isinstance(string, str):
        string = format_iterable(string)
    h = hashlib.sha256()
    h.update(bytes(string, "utf-8"))

    return str(h.hexdigest())



import importlib
import os
import time

from remotemanager.storage.sendablemixin import SendableMixin
from remotemanager.logging import LoggingMixin


class ConnectionTest(SendableMixin, LoggingMixin):
    """
    Object to store an instance of a connection test.

    Runs the tests, storing any important information

    Args:
        parent:
            parent URL
    """

    _do_not_package = ["_parent"]

    def __init__(self, parent):
        self._parent = parent

        self._run = False
        self._passed = False

        self._data = {}
        self._extra = {}

        self._latency = None

        self.dt = 0

    def __bool__(self):
        return self.passed

    @property
    def parent(self):
        return self._parent

    def exec(self) -> None:
        """
        Execute the connection test
        """

        self._run = True

        ncalls_init = self.parent.call_count
        self.dt = 0

        for name in dir(self):
            if name.startswith("test_"):
                self._logger.debug(f"running test {name}")
                self._data[name] = getattr(self, name)()

        self._passed = all(self._data.values())

        delta_calls = self.parent.call_count - ncalls_init
        latency = self.dt / delta_calls
        self._data["ncalls"] = delta_calls
        self._data["dt"] = self.dt
        self._data["latency"] = latency
        self._latency = latency

        print(f"Done! Made {delta_calls} calls, taking {self.dt:.2f}s")
        print(f"Approximate latency, {latency:.2f}s")
        if self.passed:
            print("Tests passed successfully")
        else:
            print("Tests did not pass")

    @property
    def passed(self):
        return self._passed

    @property
    def data(self):
        return self._data

    @property
    def extra(self):
        return self._extra

    @property
    def latency(self):
        return self._latency

    def test_basic(self) -> bool:
        """
        Connects to the host and returns the entry directory

        Returns:
            (bool): True if test succeeded
        """

        print("Checking for entry point...", end=" ")
        try:
            t0 = time.time()
            entry = self.parent.cmd("pwd")
            self.dt += time.time() - t0

            print(f"Success ({entry})")

            self._extra["entrypoint"] = entry
            return True

        except Exception as E:
            self._extra["entrypoint"] = str(E)
            print(f"Failure: {E}")
            return False

    def test_files(self) -> bool:
        """
        Attempts to create and delete files in several directories

        Returns:
            (bool): True if test succeeded
        """

        _ENTRYPOINT_PLACEHOLDER = "home"
        _TESTING_FILENAME = "connection_test_tmp_file"

        def create_file(directory=None):
            if directory != _ENTRYPOINT_PLACEHOLDER:
                filename = os.path.join(directory, _TESTING_FILENAME)
            else:
                filename = _TESTING_FILENAME

            t0 = time.time()
            cmd = self.parent.cmd(f"touch {filename}", raise_errors=False)
            presence = (
                _TESTING_FILENAME
                in self.parent.cmd(f"ls {filename}", raise_errors=False).stdout
            )
            dt = time.time() - t0

            return cmd, presence, filename, dt

        dirs = {_ENTRYPOINT_PLACEHOLDER: False, "/tmp": False, "/scratch": False}

        for directory in dirs:
            print(f"Checking file creation in {directory}...", end=" ")
            cmd, presence, created, dt = create_file(directory)
            dirs[directory] = presence
            print(presence)

            if presence:
                t0 = time.time()
                self.parent.cmd(f"rm {created}")
                self.dt += time.time() - t0
            else:
                self._extra[f"{directory} creation error"] = cmd.stderr

        self._extra["file_creation"] = dirs

        return dirs[_ENTRYPOINT_PLACEHOLDER]

    def test_transport(self) -> bool:
        """
        Tests all available transport methods, returning True if at least one
        is functional

        Returns:
            (bool): True if test succeeded
        """

        transport_collection = importlib.import_module("remotemanager.transport")

        transport_instances = []
        for name in dir(transport_collection):
            # exclude private/protected variables
            # exclude base class and cp (local only)
            if name.startswith("_") or name in ("transport", "cp"):
                continue
            # generate and store an instance
            transport_instances.append(transport_collection.__dict__[name](self.parent))

        written_files = []
        if len(transport_instances) > 0:
            for i in range(3):
                tmp = f"test_file_{i}.txt"
                with open(tmp, "w+") as o:
                    o.write(f"test_{i}")
                    written_files.append(tmp)

        passed = False
        transport_tests = {}
        for tmp_transport in transport_instances:
            name = tmp_transport.__module__
            print(f"Testing {name}:\n\tsend... ", end="")

            t0 = time.time()
            self.parent.cmd("mkdir -p connection_test")

            tmp_transport.queue_for_push(written_files, ".", "connection_test")
            try:
                tmp_transport.transfer()
                send = self.parent.utils.ls("connection_test") == written_files
            except Exception as E:
                send = False
                self._extra[f"{name} send error"] = str(E)

            print(f"{send}\n\tpull... ", end="")

            self.parent.cmd("touch connection_test/retrieve_test")

            tmp_transport.queue_for_pull(["retrieve_test"], ".", "connection_test")
            try:
                tmp_transport.transfer()
                pull = "retrieve_test" in self.parent.utils.ls(".", local=True)
            except Exception as E:
                pull = False
                self._extra[f"{name} pull error"] = str(E)

            print(pull)

            self.parent.cmd("rm -r connection_test")
            self.dt += time.time() - t0
            try:
                os.remove("retrieve_test")
            except FileNotFoundError:
                pass

            transport_tests[tmp_transport.__module__] = (send, pull)

            if send and pull:
                passed = True

        self._extra["transport tests"] = transport_tests

        print("Cleaning up... ", end="")
        for file in written_files:
            os.remove(file)
        print("Done")

        return passed



import re


def detect_locale_error(stderr):
    """
    Given a stderr output string `stderr`, will regex search for the locale errors

    Args:
        stderr:
            stderr text
    Returns:
        bool: True if the error appears to be a locale issue
    """
    match = r"(?i)(?:set|setting)\s?locale"

    return len(re.findall(match, stderr)) != 0



from remotemanager.connection.computers.utils import format_time
from remotemanager.connection.computers.base import BaseComputer
from remotemanager.connection.computers.resource import Resource, Resources

__all__ = ["format_time", "BaseComputer", "Resource", "Resources"]



from typing import Union


def format_time(time: Union[int, str]) -> str:
    """
    Take integer seconds and generate a HH:MM:SS timestring

    Args:
        time (int):
            seconds
    Returns:
        (str):
            HH:MM:SS format timestamp
    """
    if not isinstance(time, int):
        # if given a string, assume format is already valid and return
        return time
    mins = time // 60
    hours = mins // 60

    secstring = str(time % 60).rjust(2, "0")
    minstring = str(mins % 60).rjust(2, "0")
    hourstring = str(hours).rjust(2, "0")

    return f"{hourstring}:{minstring}:{secstring}"


def time_to_s(time: str) -> int:
    """Convert back from HH:MM:SS to integer seconds"""
    hh, mm, ss = time.split(":")

    return int(hh) * 3600 + int(mm) * 60 + int(ss)



from remotemanager.connection.computers import Resources


def slurm(resources: Resources) -> list:
    """
    Example parser for a slurm based computer

    Args:
        resources:
            resources dictionary, provided by BaseComputer
    Returns:
        list of resource request lines
    """
    from remotemanager.connection.computers import format_time

    output = []
    for resource in resources:
        if resource.name == "time":
            formatted = format_time(resource.value)
            output.append(f"#SBATCH --{resource.flag}={formatted}")

        elif resource:
            output.append(f"#SBATCH {resource.flag} {resource.value}")

    return output


def torque(resources: Resources) -> list:
    """
    Example parser for a torque based computer

    Args:
        resources:
            resources dictionary, provided by BaseComputer
    Returns:
        list of resource request lines
    """
    from remotemanager.connection.computers import format_time

    output = []
    for resource in resources:
        if resource.name in ["mpi", "omp", "nodes", "time"]:
            continue
        elif resource:
            output.append(f"#PBS -{resource.flag} {resource.value}")
    # this sort is unnecessary in a real parser, it's just here to help the CI tests
    output = sorted(output)

    output.append(
        f"#PBS -l nodes={resources['nodes'].value}:"
        f"ppn={resources['mpi'].value},"
        f"walltime={format_time(resources['time'].value)}"
    )

    output.append("\ncd $PBS_O_WORKDIR")
    output.append(f"export OMP_NUM_THREADS={resources['omp']}")

    return output



import os
import subprocess
import getpass
import time
import warnings

from remotemanager.utils.uuid import generate_uuid
from remotemanager.connection.detect_locale_error import detect_locale_error
from remotemanager.storage.sendablemixin import SendableMixin, Unloaded
from remotemanager.logging import LoggingMixin


def _process_redirect_file(file) -> [str, None]:
    if file is not None:
        return os.path.abspath(file)
    return None


def _clean_output(output) -> [str, None]:
    """
    Wrapper for the string.strip() method, allowing None

    Args:
        output:
            string (or None) to be handled

    Returns (str, None):
        cleaned cmd output
    """
    if output is None:
        return None
    return output.strip()


class CMD(SendableMixin, LoggingMixin):
    """
    This class stores a command to be executed, and the returned stdout, stderr

    Args:
        cmd (str):
            command to be executed
        asynchronous (bool):
            execute commands asynchronously
            defaults to False
        stdout (str):
            optional file to redirect stdout to
        stderr (str):
            optional file to redirect stderr to
        timeout (int):
            time to wait before issuing a timeout
        max_timeouts (int):
            number of times to attempt communication in case of a timeout
        force_file (bool):
            always use the fexec method if True
    """

    _do_not_package = ["_subprocess"]

    def __init__(
        self,
        cmd: str,
        asynchronous: bool = False,
        stdout: str = None,
        stderr: str = None,
        timeout: int = 5,
        max_timeouts: int = 3,
        raise_errors: bool = True,
        force_file: bool = False,
    ):
        self._uuid = generate_uuid(f"{time.time()} {cmd}")
        # command to execute
        self._cmd = cmd
        # settings
        self._async = asynchronous
        # force file-based exec
        self._force_file = force_file
        # stdout/stderr redirect
        if stderr is not None and stderr == stdout:
            raise_errors = False
            warnings.warn(
                "stderr and stdout are pointed at the same file, "
                "this will cause errors to be suppressed"
            )
        self._redirect = {
            "stdout": _process_redirect_file(stdout),
            "stderr": _process_redirect_file(stderr),
            "execfile": None,
        }

        if not asynchronous:
            initmsg = "creating a new CMD instance"
        else:
            initmsg = "creating a new asynchronous CMD instance"

        self._logger.info(initmsg)

        # timeout info
        self.timeout = timeout
        self.max_timeouts = max_timeouts
        self._timeout_current_tries = 0

        # call duration storage
        self._duration = {}

        # prefer to raise an error, or continue
        self._raise_errors = raise_errors

        # supplementary data for post-exec
        self._subprocess = None
        self._cached = False
        self._stdout = None
        self._stderr = None
        self._returncode = None
        self._whoami = None
        self._pwd = None
        self._pid = None

    def __repr__(self):
        stdout = self.stdout if self.stdout is not None else ""
        return stdout

    @property
    def uuid(self):
        return self._uuid

    @property
    def short_uuid(self):
        return self._uuid[:8]

    @property
    def tempfile(self):
        return f"{self.short_uuid}.sh"

    @property
    def sent(self) -> str:
        """
        The command passed at initialisation
        """
        return self._cmd.__repr__()

    @property
    def asynchronous(self) -> bool:
        """
        True if commands are to be executed asynchronously
        """
        return self._async

    @property
    def is_redirected(self) -> bool:
        """
        True if the cmd is redirected to a file
        """
        return any([self._redirect["stdout"], self._redirect["stderr"]])

    def _get_return_attr(self, attr: str) -> [str, None]:
        # subprocess objects cannot be serialised, check if we are post-reserialisation
        if isinstance(self._subprocess, Unloaded):
            self._logger.warning(f"broken subprocess, getting attr _{attr}")
            return getattr(self, f"_{attr}")

        return self.communicate()[attr]

    @property
    def stdout(self) -> str:
        """
        Directly returns the stdout from the cmd execution. Attempts
        to communicate with the subprocess in the case of an async run.

        Returns None if the command has not been executed yet.

        Returns (str):
            the stdout from the command execution
        """
        if self._stdout is not None:
            self._logger.info("returning cached stdout")
            return self._stdout
        self._stdout = self._get_return_attr("stdout")
        return self._stdout

    @property
    def stderr(self) -> str:
        """
        Directly returns the stderr from the cmd execution. Attempts
        to communicate with the subprocess in the case of an async run.

        Returns None if the command has not been executed yet.

        Returns (str):
            the stdout from the command execution
        """
        if self._stderr is not None:
            self._logger.info("returning cached stderr")
            return self._stderr
        self._stderr = self._get_return_attr("stderr")
        return self._stderr

    @property
    def pwd(self) -> str:
        """
        Present working directory at command execution

        Returns None if the command has not been executed yet.

        Returns (str):
            working dir of command execution
        """
        return self._pwd

    @property
    def whoami(self) -> str:
        """
        Present user at command execution

        Returns None if the command has not been executed yet.

        Returns (str):
            username who executed the command
        """
        return self._whoami

    @property
    def pid(self) -> int:
        """
        The Process ID of the spawned process

        Returns None if the command has not been executed yet.

        Returns (int):
            the PID of the spawned shell for this command
        """
        return self._pid

    @property
    def returncode(self) -> [int, None]:
        """
        Attempt to retrieve the returncode of the subprocess. This call will
        not disturb an asynchronous run, returning None

        Returns (int, None):
                The exit status of the subprocess, None if it is still running.
                None otherwise.
        """
        if self._subprocess is not None:
            self._subprocess.poll()
            self._returncode = self._subprocess.returncode
        return self._returncode

    @property
    def is_finished(self) -> bool:
        """
        Returns True if this command has finished execution. This will NOT talk
        to the process, as to not disturb async runs, so will always return
        False in those instances

        Returns (bool):
                True if the command has completed
        """
        return self.returncode is not None

    @property
    def succeeded(self) -> [None, bool]:
        """
        True if the command successfully executed

        Returns:
            None if not finished, True if returncode is 0
        """
        if not self.is_finished:
            return None
        return self.returncode == 0

    @property
    def duration(self):
        return self._duration

    @property
    def latency(self):
        return sum(self.duration.values())

    def exec(self) -> None:
        """
        Executes the command, storing execution info and in the
        case of a non-async run; returned values

        Returns:
            None
        """
        self._whoami = getpass.getuser()
        self._pwd = os.getcwd()

        if self.is_redirected:
            out = self._redirect["stdout"]
            err = self._redirect["stderr"]
            stdout = open(out, "w+") if out is not None else None
            stderr = open(err, "w+") if err is not None else None
        else:
            stdout = subprocess.PIPE
            stderr = subprocess.PIPE

        if self._force_file:
            return self._fexec(stdout, stderr)

        try:
            self._exec(stdout, stderr)
        except OSError as E:
            msg = "Encountered an OSError on exec, attempting file exec"
            warnings.warn(msg)
            self._logger.warning(E)
            self._logger.warning(msg)
            self._fexec(stdout, stderr)

    def _exec(self, stdout, stderr) -> None:
        """
        Directly executes the command

        Args:
            stdout:
                stdout passthrough
            stderr:
                stderr passthrough

        Returns:
            None
        """
        self._logger.debug(f"executing command in {self.pwd}")
        self._logger.debug(f'"{self._cmd}"')

        hostexec = "/bin/bash" if os.name != "nt" else None

        t0 = time.time()
        self._subprocess = subprocess.Popen(
            self._cmd,
            stdout=stdout,
            stderr=stderr,
            shell=True,
            text=True,
            executable=hostexec,
        )
        self._duration["exec"] = time.time() - t0
        self._pid = self._subprocess.pid
        if not self._async and not self.is_redirected:
            self._logger.debug("in-exec communication triggered")
            self.communicate()

    def _fexec(self, stdout, stderr) -> None:
        """
        Executes the command by first writing it to a file

        Args:
            stdout:
                stdout passthrough
            stderr:
                stderr passthrough

        Returns:
            None
        """
        file = self.tempfile
        with open(file, "w+") as o:
            o.write(self._cmd)

        # noinspection PyTypedDict
        self._redirect["execfile"] = file

        t0 = time.time()
        self._subprocess = subprocess.Popen(
            f"bash {file}",
            stdout=stdout,
            stderr=stderr,
            shell=True,
            universal_newlines=True,
            executable="/bin/bash",
        )
        self._duration["exec"] = time.time() - t0
        self._pid = self._subprocess.pid

        if not self._async and not self.is_redirected:
            self._logger.debug("in-exec communication triggered")
            self.communicate()

    def communicate(self, use_cache: bool = True, ignore_errors: bool = False) -> dict:
        """
        Communicates with the subprocess, returning the stdout and stderr in
        a dict

        Args:
            use_cache (bool):
                use cached value if it is available
            ignore_errors (bool):
                do not raise error regardless of base setting

        Returns (dict):
                {'stdout': stdout, 'stderr': stderr}
        """
        if self._cached and use_cache:
            self._logger.info("using cached return values")
            return {
                "stdout": self._stdout,
                "stderr": self._stderr,
            }
        elif not self.is_redirected:
            self._logger.info(f"communicating with process {self.pid}")
            std, err = self._communicate()
        else:
            self._logger.info("files are redirected, attempting to read")
            std, err = self._file_communicate()

        if std is not None:
            self._stdout = std.strip()
        if err is not None:
            self._stderr = err.strip()

        def format_output(string):
            if string is None:
                return

            if len(string.split("\n")) <= 1:
                return string

            return "\n".join([f"  {line}" for line in string.split("\n")])

        self._logger.info(f"stdout from exec: |\n{format_output(std)}")
        if err:
            self._logger.warning(f"stderr from exec: |\n{format_output(err)}")

        if std or err:  # skip if all None
            self._logger.debug("caching results")
            self._cached = True

        if self._redirect["execfile"] is not None:
            tf = self._redirect["execfile"]
            try:
                # noinspection PyTypeChecker
                os.remove(tf)
                self._redirect["execfile"] = None
                self._logger.info(f"removed temp file {tf}")
            except FileNotFoundError:
                self._logger.info(f"temp file {tf} not found")
                pass

        if self._raise_errors and not ignore_errors and err is not None and err != "":
            if detect_locale_error(err):
                self._logger.warning(f"locale error detected: {err}")
            else:
                raise RuntimeError(f"received the following stderr: \n{err}")

        self._stdout = _clean_output(std)
        self._stderr = _clean_output(err)

        if self._stderr is not None and self._stderr == "" and self.returncode != 0:
            warnings.warn(
                f"stderr is empty, but return code != 0 ({self.returncode}). "
                f"This could indicate an error."
            )

        return {"stdout": self._stdout, "stderr": self._stderr}

    def _communicate(self) -> (str, str):
        """
        Attempt to communicate with the process, handling timeout

        Issues a Popen.communicate() with a timeout
        If this fails, will wait for (timeout * number of tries) seconds and
        try again. This continues until max_timeouts has been reached

        Returns (str, str):
            stdout, stderr
        """
        timeout = self.timeout
        self._timeout_current_tries += 1
        dt = 0  # accumulate time on each retry, rather than the whole call
        try:
            t0 = time.time()
            stdout, stderr = self._subprocess.communicate(timeout=timeout)
            dt += time.time() - t0
        except AttributeError as E:
            self._logger.info("attempted communicate on unexec CMD")
            self._logger.info(str(E))
            return None, None
        except subprocess.TimeoutExpired:
            print(
                f"({self._timeout_current_tries}/{self.max_timeouts}) "
                f"communication attempt failed after {timeout}s",
                end="... ",
            )

            if (
                isinstance(self.max_timeouts, int)
                and 0 < self.max_timeouts <= self._timeout_current_tries
            ):
                print("could not communicate, killing for safety")
                self.kill()
                raise RuntimeError("could not communicate with process")

            waittime = timeout * self._timeout_current_tries

            print(f"waiting {waittime}s and trying again")
            time.sleep(waittime)

            return self._communicate()

        self._duration["communicate"] = dt
        return stdout, stderr

    def _file_communicate(self):
        """
        We are redirected to a file, attempt to read the output
        """
        self._subprocess.poll()
        returncode = self._subprocess.returncode
        count = 0
        while returncode is None:
            time.sleep(0.05)
            self._subprocess.poll()
            returncode = self._subprocess.returncode
            count += 1
            if count >= 10:
                raise RuntimeError("could not communicate with process")

        outfile = self._redirect["stdout"]
        errfile = self._redirect["stderr"]

        if outfile is not None:
            self._logger.debug(f"reading file {outfile}")
            with open(outfile, "r") as o:
                std = o.read().strip()
        else:
            self._logger.debug("outfile is None")
            std = None

        if errfile is not None:
            self._logger.debug(f"reading file {errfile}")
            with open(errfile, "r") as e:
                err = e.read().strip()
        else:
            self._logger.debug("errfile is None")
            err = None

        return std, err

    def kill(self) -> None:
        """
        Kill the process associated with this command, if one exists

        Returns:
            None
        """
        self._logger.info("received termination call")
        if self._subprocess is not None:
            self._logger.debug("_subprocess exists, sending kill()")
            self._subprocess.kill()
            self._logger.debug("polling process...")
            self._subprocess.poll()
            self._logger.debug("polling complete")
            return None
        self._logger.debug("process has not been launched yet")



from remotemanager.logging import Quiet


class LazyAppend:
    """
    Object which enables a lazy_append context manager

    """

    def __init__(self, parent):
        self._parent = parent

    def __enter__(self, *args, **kwargs):
        """
        When targeting a variable (``with ds.lazy_append() as var:``),

        `var` is what is returned from `__enter__`

        Thus, we need a copy of the class
        """
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        if self._parent.dependency is not None:
            for ds in self._parent.dependency.ds_list:
                ds.finish_append()
            # Since Quiet is global, setting it to False within the finish_append will
            # cause it to be also False in the subsequent calls
            Quiet.state = False
        else:
            self._parent.finish_append()

    def append_run(self, *args, **kwargs):
        kwargs["lazy"] = True
        self._parent.append_run(*args, **kwargs)



"""
DynamicValue stub class allows for deferred calculation of values.

Constructing a "tree" of values using these objects allows for later
assessment. Used in Computers for dynamic resource assignment.

>>> val_a = DynamicValue(10)
>>> val_b = DynamicValue(6)
>>> val_c = DynamicValue(val_a + val_b)
>>> val_c.value
16
"""

from numbers import Number
from typing import Union

from remotemanager.storage import SendableMixin


class DynamicValue(SendableMixin):
    """
    Args:
        a:
            "First" number in operation
        b:
            "Second" number in operation. Can be None,
            in which case this value is considered "toplevel"
        op:
            Operation to use. Can be None for toplevel values
        default:
            Default value can be set in case the primary value is
            set to None
    """

    __slots__ = ["_a", "_b", "_op", "_default"]

    def __init__(
        self,
        a: Union[Number, "DynamicValue", None],
        b: Union[Number, "DynamicValue", None] = None,
        op: Union[str, None] = None,
        default: Union[Number, None] = None,
    ):
        if b is None and op is not None:
            raise ValueError("Operator specified without 2nd value")
        if b is not None and op is None:
            raise ValueError("Cannot specify 2nd value without operator")
        self._a = a
        self._b = b
        self._op = op
        self._default = default

    def __pow__(self, other: Union[Number, "DynamicValue"]) -> "DynamicValue":
        obj = DynamicValue(self, other, "pow")
        return obj

    def __truediv__(self, other: Union[Number, "DynamicValue"]) -> "DynamicValue":
        obj = DynamicValue(self, other, "div")
        return obj

    def __mul__(self, other: Union[Number, "DynamicValue"]) -> "DynamicValue":
        obj = DynamicValue(self, other, "mul")
        return obj

    def __add__(self, other: Union[Number, "DynamicValue"]) -> "DynamicValue":
        obj = DynamicValue(self, other, "add")
        return obj

    def __sub__(self, other: Union[Number, "DynamicValue"]) -> "DynamicValue":
        obj = DynamicValue(self, other, "sub")
        return obj

    def __repr__(self):
        ops = {"pow": "**", "div": "/", "mul": "*", "add": "+", "sub": "-"}
        try:
            op = ops[self.op]
        except KeyError:
            return str(self._a)
        return f"DynamicValue({self._a}{op}{self._b})"

    @property
    def a(self):
        """
        Returns:
            Value of "first" number
        """
        if isinstance(self._a, DynamicValue):
            return self._a.value
        return self._a

    @property
    def b(self):
        """
        Returns:
            Value of "second" number
        """
        if isinstance(self._b, DynamicValue):
            return self._b.value
        return self._b

    @property
    def op(self):
        """
        Returns:
            Operation string
        """
        return self._op

    @property
    def default(self):
        """
        Returns:
            The default value
        """
        return self._default

    @default.setter
    def default(self, default):
        """default setter"""
        self._default = default

    @property
    def value(self):
        """
        Calculates value by calling the whole chain of numbers

        Returns:
            Value
        """
        if self.op is None and self.a is not None:
            return self.a
        elif self.b is None:
            return self.default

        if self.op == "pow":
            return self.a**self.b
        if self.op == "div":
            return self.a / self.b
        if self.op == "mul":
            return self.a * self.b
        if self.op == "add":
            return self.a + self.b
        if self.op == "sub":
            return self.a - self.b

    @value.setter
    def value(self, val):
        """
        It is possible to update the value of a toplevel DynamicValue

        Args:
            val:
                New Value
        """
        if isinstance(self._a, DynamicValue):
            raise RuntimeError("Can only set values at toplevel, e.g. DynamicValue(x)")
        self._a = val



import copy
import json
import typing
import importlib
import warnings

import yaml

from typing import Union

from remotemanager.connection.url import URL
from remotemanager.connection.computers.resource import Resource, Resources
from remotemanager.storage.function import Function
from remotemanager.storage.sendablemixin import get_class_storage, INTERNAL_STORAGE_KEYS


class BaseComputer(URL):
    """
    Base computer module for HPC connection management.

    Extend this class for connecting to your machine
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self._extra = ""
        self._internal_extra = ""

    def __setattr__(self, key, value):
        """
        If the set `key` attribute is a Resource, and `value` is not a Resource,
        instead set the `value` of that attribute

        Args:
            key:
                attribute name to set
            value:
                value to set to

        Returns:
            None
        """
        # if this resource already exists, and we're _not_ trying to set a new resource,
        # set the value of the Resource to value
        if key in self.__dict__ and not isinstance(value, Resource):
            if isinstance(getattr(self, key), Resource):
                # need to get before we can set
                getattr(self, key).value = value
                return

        object.__setattr__(self, key, value)

    def get_parser(self):
        parser = getattr(self, "parser", None)

        if parser is None:
            # legacy _parser
            parser = getattr(self, "_parser", None)

        return parser

    @classmethod
    def from_dict(cls, spec: dict, **url_args):
        """
        Create a Computer class from a `spec` dictionary. The required values are:

        - resources:
            a dict of required resources for the machine (mpi, nodes, queue, etc.)
        - resource_parser:
            a function which takes a dictionary of {resource: Option}, returning a list
            of valid jobscript lines

        You can also provide some optional arguments:

        - required_or:
            list of resources, ONE of which is required. Note that this must be a
            _list_ of dicts, one for each or "block"
        - optional_resources:
            as with `resources`, but these will not be stored as required values
        - optional_defaults:
            provide defaults for the names given in optional_resources. When adding the
            optional arg, the optional_defaults will be checked to see if a default is
            provided
        - host:
            machine hostname. Note that this _must_ be available for a job run, but if
            not provided within the spec, can be added later with
            ``url.host = 'hostname'``
        - submitter:
            override the default submitter
        - python:
            override the default python
        - extra:
            any extra lines that should be appended after the resource specification.
            Note that this includes module loads/swaps, but this can be specified on a
            per-job basis, rather than locking it into the `Computer`

        The `resources` specification is in `notebook`:`machine` order. That is to say
        that the `key` is what will be required in the _notebook_, and the `value` is
        what is placed in the jobscript:

        >>> spec = {'resources': {'mpi': 'ntasks'}, ...}
        >>> url = BaseComputer.from_dict(spec)
        >>> url.mpi = 12
        >>> url.script()
        >>> "--ntasks=12"

        Args:
            spec (dict):
                input dictionary
            url_args:
                any arguments to be passed directly to the created `url`

        Returns:
            Computer class as per input spec
        """
        from remotemanager import Logger

        payload = copy.deepcopy(spec)
        # check if we have a stored Computer Class
        class_store_key = INTERNAL_STORAGE_KEYS["CLASS_STORAGE_KEY"]
        if class_store_key in payload:
            class_storage = payload[class_store_key]
            Logger.debug(f"using class storage {class_storage}")
            # first, try grab the module
            try:
                mod = importlib.import_module(class_storage["mod"])
                # then try import from that module
                try:
                    cls = getattr(mod, class_storage["name"])
                except AttributeError:  # inner import error, no class
                    warnings.warn(f"Could not import class {class_storage['name']}")
            except ModuleNotFoundError:  # outer import error, no module
                warnings.warn(f"Could not import module {class_storage['mod']}")

        Logger.debug(f"from_dict called on {cls}")

        # convert the parser back into a function
        def unpack(parser_data) -> Function:
            """
            Parser data can be either a string, dict or callable.
            Handle all and return the Function

            Args:
                parser_data:
                    stored parser
            Returns:
                Function
            """
            if callable(parser_data):
                return Function(parser_data, force_self=True)
            if isinstance(parser_data, str):
                return Function(parser_data, force_self=True)
            if isinstance(parser_data, dict):
                return Function.unpack(parser_data, force_self=True)
            if isinstance(parser_data, Function):
                return parser_data

            Logger.warning(
                f"no string found, setting {type(parser_source)} "
                f"{parser_source} as parser"
            )
            return parser_data

        parser = None
        parser_source = payload.pop("resource_parser_source", None)
        if parser_source is not None:
            Logger.debug("generating parser from source code")
            parser = unpack(parser_source)
        else:  # legacy unpack method
            Logger.debug("parser source code not found at resource_parser_source")
            parser_source = payload.pop("resource_parser", None)
            parser = unpack(parser_source)

        # parser object must be assigned (monkey patched) at class level, or it will
        # not be a bound-method, which is required for `self` and inspect to work
        if parser is not None:
            Logger.debug("assigning parser object at parser")
            cls.parser = parser.object
            cls._parser_source = parser.raw_source
        else:
            Logger.debug("parser NOT assigned")

        # create a new class
        computer = cls()
        if parser is not None:
            computer._parser_source = parser.raw_source

        # add the resource storing objects to the class
        required_resources = payload.pop("resources")
        oldstyle = False
        for field, resource_args in required_resources.items():
            if isinstance(resource_args, str):
                # this is a consequence of an older style "name": "flag" spec
                # cast to new type by setting it to "flag"
                # other args can be set later
                oldstyle = True
                resource_args = {"flag": resource_args, "optional": False}

            resource = Resource(name=field, **resource_args)
            setattr(computer, field, resource)

        if oldstyle:
            optional = payload.pop("optional_resources", {})
            for name, flag in optional.items():
                resource = Resource(name=name, flag=flag)
                setattr(computer, name, resource)

            required_or = payload.pop("required_or", [])
            for group in required_or:
                for name, flag in group.items():
                    replaces = [n for n in group if n != name]
                    resource = Resource(
                        name=name, flag=flag, optional=False, replaces=replaces
                    )
                    setattr(computer, name, resource)

            defaults = payload.pop("optional_defaults")
            for name, default in defaults.items():
                computer.resource_dict[name].default = default

        if oldstyle:
            print(
                "WARNING! Old style import detected. "
                "You should check the validity of the "
                "resulting Computer then re-dump to yaml."
            )

        # add any extra content
        for key, val in payload.items():
            setattr(computer, key, val)

        for key, val in url_args.items():
            setattr(computer, key, val)

        return computer

    def to_dict(self, include_extra: bool = True) -> dict:
        """
        Generate a spec dict from this Computer

        Args:
            include_extra:
                includes the `extra` property if True (default True)

        Returns:
            dict
        """
        self._logger.debug(f"to_dict called on {self}")

        def create_entry(obj):
            data = {"flag": obj.flag}

            if obj.min is not None:
                data["min"] = obj.min
            if obj.max is not None:
                data["max"] = obj.max
            if getattr(obj, "default", None) is not None:
                data["default"] = obj.default
            if not obj.optional:
                data["optional"] = False
            if len(obj.requires) != 0:
                data["requires"] = obj.requires
            if len(obj.replaces) != 0:
                data["replaces"] = obj.replaces
            if obj.format is not None:
                data["format"] = obj.format

            return data

        # gather all non resource objects
        spec = {
            k: getattr(self, k)
            for k in self.__dict__
            if k not in self.resources and not k.startswith("_")
        }

        spec["resources"] = {n: create_entry(r) for n, r in self.resource_dict.items()}

        if getattr(self, "_parser_source", None) is not None:
            parser = Function(self._parser_source, force_self=True).raw_source
        else:
            parser = self.get_parser()
            try:
                # avoids a strange error where the parser source cant be found.
                # likely loads it into memory where `inspect` can access it
                # noinspection PyStatementEffect
                parser
            except AttributeError:
                pass

        if isinstance(parser, Function):
            spec["resource_parser_source"] = parser.raw_source
        elif parser is not None:
            spec["resource_parser_source"] = Function(
                parser, force_self=True
            ).raw_source
        # grab `extra` if requested
        if include_extra and self.extra is not None:
            spec["extra"] = self.extra
        # round up missing package inclusions
        collect = ["submitter", "shebang", "pragma"]
        for name in collect:
            try:
                spec[name] = getattr(self, name)
            except AttributeError:
                pass
        # parser object is not required, and looks ugly
        if "parser" in spec:
            del spec["parser"]
        cls_store_key = INTERNAL_STORAGE_KEYS["CLASS_STORAGE_KEY"]
        spec[cls_store_key] = get_class_storage(self)
        return spec

    @classmethod
    def from_yaml(cls, filepath: str, **url_args):
        """
        Create a Computer from `filepath`.

        Args:
            filepath:
                path containing yaml computer spec
            **url_args:
                extra args to be passed to the internal URL

        Returns:
            BaseComputer
        """
        if isinstance(filepath, str):
            try:
                with open(filepath, "r") as o:
                    data = yaml.safe_load(o)
            except OSError:
                data = yaml.safe_load(filepath)
        else:
            data = yaml.safe_load(filepath)

        return cls.from_dict(data, **url_args)

    def to_yaml(
        self, filepath: Union[str, typing.IO, None] = None, include_extra: bool = True
    ) -> Union[str, None]:
        """
        Dump a computer to yaml `filepath`.

        Args:
            filepath:
                path containing yaml computer spec
            include_extra:
                includes the `extra` property if True (default True)
        """
        data = self.to_dict(include_extra)
        # source will simply not print correctly with base yaml
        # extract it and do it manually, if it exists
        if "resource_parser_source" in data:
            parser_string = ["resource_parser_source: |"] + [
                f"    {line}" for line in data.pop("resource_parser_source").split("\n")
            ]
            # dump the remaining content to string
            prepared = yaml.dump(data)
            # append the cleaned string
            prepared += "\n".join(parser_string)
        else:
            prepared = yaml.dump(data)

        if filepath is None:  # "dump" to string
            return prepared
        elif isinstance(filepath, str):  # dump to path
            with open(filepath, "w+") as o:
                o.write(prepared)
        else:  # assume file handler and dump there
            filepath.write(prepared)

    @classmethod
    def from_repo(
        cls,
        name: str,
        branch: str = "main",
        repo: str = "https://gitlab.com/l_sim/remotemanager-computers/",
        **url_args,
    ):
        """
        Attempt to access the remote-computers repo, and pull the computer with name
        `name`

        Args:
            name (str):
                computer name to target
            branch (str):
                repo branch (defaults to main)
            repo (str):
                repo web address (defaults to main l_sim repo)

        Returns:
            BaseComputer instance
        """
        import requests
        from remotemanager.utils import ensure_filetype

        def download_file(file_url, filename):
            response = requests.get(file_url)

            if response.status_code == requests.codes.ok:
                # Save the file
                with open(filename, "wb") as file:
                    file.write(response.content)
                print(f"Grabbed file '{filename}'")
            else:
                raise RuntimeError(f"Could not find a file at: {file_url}")

        filename = ensure_filetype(name, "yaml").lower()
        url = f"{repo}-/raw/{branch}/storage/{filename}"

        print(f"polling url {url}")

        download_file(url, filename)

        return cls.from_yaml(filename, **url_args)

    def generate_cell(
        self, name: Union[str, None] = None, return_string: bool = False
    ) -> Union[None, str]:
        """
        Prints out copyable source which regenerates this Computer

        Args:
            name (str, None):
                Optional name for new computer. Defaults to `new`
            return_string (bool):
                Also returns the string if True. Defaults to False

        Returns:
            (None, str)
        """
        if name is None:
            name = "new"
        output = [
            "# Copy the following into a jupyter cell or python script "
            "to generate a modifiable source",
            "\n# Parser source code",
        ]
        source = self.to_dict()

        try:
            parser = self.get_parser()
            if not isinstance(parser, Function):
                parser = Function(parser)

            output.append(parser.raw_source)

            source.pop("resource_parser_source")
            # use json.dumps with indent=4 to format dict
            output.append(
                f"\n# JSON compatibility\n"
                f"true = True\n"
                f"false = False\n\n"
                f"# spec dict\n\nspec = {json.dumps(source, indent=4)}"
            )
            output.append(f'spec["resource_parser"] = {parser.name}')

            output.append(f"\n{name} = BaseComputer.from_dict(spec)")

        except TypeError:
            output.append(
                f"# JSON compatibility\n"
                f"true = True\n"
                f"false = False\n\n"
                f"# spec dict\n\nspec = {json.dumps(source, indent=4)}"
            )
            output.append(f"\n{name} = BaseComputer.from_dict(spec)")

        output = "\n".join(output)

        print(output)

        if return_string:
            return output

    @property
    def resources(self) -> list:
        return sorted([k for k, v in self.__dict__.items() if isinstance(v, Resource)])

    @property
    def resource_objects(self) -> list:
        return [v for k, v in self.__dict__.items() if isinstance(v, Resource)]

    @property
    def resource_dict(self) -> dict:
        return {k.strip(): getattr(self, k) for k in self.resources}

    @property
    def required(self) -> list:
        """
        Returns a list of required arguments
        """
        required = []

        def append_if(item):
            if item not in required:
                required.append(item)

        for name, resource in self.resource_dict.items():
            if not resource.optional:
                append_if(name)
            for name in resource.requires:
                append_if(name)

        return required

    @property
    def missing(self) -> list:
        """
        Returns the currently missing arguments
        """
        missing = []
        covered = []
        for resource in self.resource_objects:
            # this resource has a value, so is not missing, and can replace others
            if resource:
                covered.append(resource.name)
                for name in resource.replaces:
                    if name in missing:
                        missing.remove(name)
                    covered.append(name)
            # resource is missing a value and is non optional
            elif not resource and not resource.optional:
                if resource.name not in covered:
                    missing.append(resource.name)

            for name in resource.requires:
                if name not in covered and not self.resource_dict[name]:
                    missing.append(name)

        return missing

    @property
    def valid(self):
        return len(self.missing) == 0

    def update_resources(self, **kwargs):
        """
        Set any arguments passed to the script call

        Args:
            **kwargs:
                kwarg updates e.g. mpi=128
        Returns:
            None
        """
        for k, v in kwargs.items():
            setattr(self, k, v)

    def parser(self, resources) -> list:
        """
        Default parser for use on basic "SLURM style" machines.

        Will iterate over resource objects, creating a script of the format:

        {pragma} --{flag}={value}

        ..note::
            This method can (and should) be overidden for a custom parser.

        Args:
            resources:
                Resources object, to be created by BaseComputer

        Returns:
            list of resource lines
        """
        output = []
        for r in resources:
            if r:
                output.append(r.resource_line)

        return output

    @property
    def extra(self):
        return self._internal_extra + self._extra

    @extra.setter
    def extra(self, external):
        self._extra = external

    def script(self, **kwargs) -> str:
        """
        Takes job arguments and produces a valid jobscript

        Returns:
            (str):
                script
        """
        self.update_resources(**kwargs)
        if not self.valid:
            raise RuntimeError(f"missing required arguments: {self.missing}")

        pragma = getattr(self, "pragma", None)
        tag = getattr(self, "resource_tag", None)
        sep = getattr(self, "resource_separator", None)
        submit_args = Resources(
            resources=self.resource_objects,
            pragma=pragma,
            tag=tag,
            separator=sep,
            run_args=kwargs,
        )

        script = [self.shebang]

        script += self.parser(submit_args)

        if self.extra is not None:
            script += [self.extra]

        return "\n".join(script)

    def pack(self, file=None):
        if file is not None:
            self.to_yaml(filepath=file)
            return
        return self.to_dict()

    @classmethod
    def unpack(cls, data: dict = None, file: str = None, limit: bool = True):
        if file is not None:
            return cls.from_yaml(file)
        return cls.from_dict(data)



from remotemanager.connection.computers.base import BaseComputer
from remotemanager.connection.computers.resource import Resource


class ExampleTorque(BaseComputer):
    """
    example class for connecting to a remote computer using a torque scheduler
    """

    def __init__(self, **kwargs):
        if "host" not in kwargs:
            kwargs["host"] = "remote.address.for.connection"

        super().__init__(**kwargs)

        self.submitter = "qsub"
        self.shebang = "#!/bin/bash"
        self.pragma = "#PBS"

        self.mpi = Resource(flag="ppn", name="mpi")
        self.omp = Resource(flag="omp", name="omp")
        self.nodes = Resource(flag="nodes", name="nodes")
        self.queue = Resource(flag="q", name="queue")
        self.time = Resource(flag="walltime", name="time", format="time")
        self.account = Resource(flag="a", name="account", optional=True)
        self.jobname = Resource(flag="N", name="jobname", optional=True)
        self.outfile = Resource(flag="o", name="outfile", optional=True)
        self.errfile = Resource(flag="e", name="errfile", optional=True)

        self._internal_extra = ""

    def parser(self, resources) -> list:
        """
        Example parser for a torque based computer

        Args:
            resources:
                resources dictionary, provided by BaseComputer
        Returns:
            list of resource request lines
        """
        output = []
        for resource in resources:
            if resource.name in ["mpi", "omp", "nodes", "time"]:
                continue
            elif resource:
                output.append(f"#PBS -{resource.flag} {resource.value}")
        # this sort is unnecessary in a real parser, it's just here to help the CI tests
        output = sorted(output)

        output.append(
            f"#PBS -l nodes={resources['nodes'].value}:"
            f"ppn={resources['mpi'].value},"
            f"walltime={resources['time'].value}"
        )

        output.append("\ncd $PBS_O_WORKDIR")
        output.append(f"export OMP_NUM_THREADS={resources['omp']}")

        return output


class ExampleSlurm(BaseComputer):
    """
    example class for connecting to a remote computer using a slurm scheduler
    """

    def __init__(self, **kwargs):
        if "host" not in kwargs:
            kwargs["host"] = "remote.address.for.connection"

        super().__init__(**kwargs)

        self.submitter = "sbatch"
        self.shebang = "#!/bin/bash"
        self.pragma = "#SBATCH"

        self.mpi = Resource(flag="ntasks", name="mpi")
        self.omp = Resource(flag="cpus-per-task", name="omp")
        self.nodes = Resource(flag="nodes", name="nodes")
        self.queue = Resource(flag="queue", name="queue")
        self.account = Resource(flag="account", name="account", optional=True)
        self.time = Resource(flag="walltime", name="time", format="time")
        self.jobname = Resource(flag="job-name", name="jobname", optional=True)
        self.outfile = Resource(flag="output", name="outfile", optional=True)
        self.errfile = Resource(flag="error", name="errfile", optional=True)



"""
This module stores the placeholder arguments who's job it is to convert
arguments from the Dataset level `mpi`,  `omp`, `nodes`, etc. to what the
scheduler is expecting within a jobscript.

.. note::
    Placeholders without a `value` are "falsy". So checking their value in an
    if statement will return `True` if they have a value, `False` otherwise.
"""
import math
import warnings
from numbers import Number
from typing import Union, Any

from remotemanager.connection.computers import format_time
from remotemanager.connection.computers.dynamicvalue import DynamicValue
from remotemanager.logging import LoggingMixin
from remotemanager.storage import SendableMixin
from remotemanager.utils import ensure_list


class Resource(SendableMixin, LoggingMixin):
    """
    Stub class to sit in place of an option within a computer.

    Args:
        name (str):
            name under which this arg is stored
        flag (str):
            Flag to append value to e.g. `--nodes`, `--walltime`
        separator (str):
            Override the separator between flag and value (defaults to "=")
        tag (str):
            Override the tag preceding the flag (defaults to "--")
        default (Any, None):
            Default value, marks this Resource as optional if present
        optional (bool):
            Marks this resource as Optional. Required as there are actually
            three states:
                - Required input, required by scheduler.
                - Optional input, required by scheduler.
                - Optional input, optional by scheduler.
        requires (str, list):
            Stores the name(s) of another variable which is required alongside this one
        replaces (str, list):
            Stores the name(s) of another variable which is replaced by this one
        min (int):
            Minimum value for numeric inputs
        max (int):
            Maximum value for numeric inputs
        format (str):
            Expected format for number. Allows None, "time" or "float"
    """

    # __slots__ = [
    #     "_name",
    #     "_flag",
    #     "_value",
    #     "_requires",
    #     "_replaces",
    #     "_optional",
    #     "_min",
    #     "_max",
    #     "format"
    # ]

    def __init__(
        self,
        name: str,
        flag: Union[str, None] = None,
        tag: Union[str, None] = None,
        separator: Union[str, None] = None,
        default: Union[Any, None] = None,
        optional: bool = True,
        requires: Union[str, list, None] = None,
        replaces: Union[str, list, None] = None,
        min: Union[int, None] = None,
        max: Union[int, None] = None,
        format: Union[str, None] = None,
    ):
        self._name = name
        self._flag = flag
        self._optional = optional

        self._requires = ensure_list(requires)
        self._replaces = ensure_list(replaces)

        self.pragma = None
        self.tag = tag
        self.separator = separator

        self._value = DynamicValue(None, None, None, default)
        self.format = format

        self._min = min
        self._max = max

    def __hash__(self):
        return hash(self._flag)

    def __repr__(self):
        return str(self.value)

    def __bool__(self):
        """
        Makes objects "falsy" if no value has been set, "truthy" otherwise
        """
        return self.value is not None and self.flag is not None

    def __pow__(self, other: Union[Number, DynamicValue, "Resource"]) -> DynamicValue:
        if isinstance(other, Resource):
            other = other._value
        obj = DynamicValue(self._value, other, "pow")
        return obj

    def __mul__(self, other: Union[Number, DynamicValue, "Resource"]) -> DynamicValue:
        if isinstance(other, Resource):
            other = other._value
        obj = DynamicValue(self._value, other, "mul")
        return obj

    def __truediv__(
        self, other: Union[Number, DynamicValue, "Resource"]
    ) -> DynamicValue:
        if isinstance(other, Resource):
            other = other._value
        obj = DynamicValue(self._value, other, "div")
        return obj

    def __add__(self, other: Union[Number, DynamicValue, "Resource"]) -> DynamicValue:
        if isinstance(other, Resource):
            other = other._value
        obj = DynamicValue(self._value, other, "add")
        return obj

    def __sub__(self, other: Union[Number, DynamicValue, "Resource"]) -> DynamicValue:
        if isinstance(other, Resource):
            other = other._value
        obj = DynamicValue(self._value, other, "sub")
        return obj

    @property
    def optional(self):
        """Returns True if this Resource is optional at Dataset level"""
        return self._value.default is not None or self._optional

    @property
    def replaces(self) -> list:
        """
        List of arguments whom are no longer considered `required` if this
        resource is specified
        """
        return self._replaces

    @property
    def requires(self) -> list:
        """
        List of requirements if this resource is specified.
        e.g. nodes for mpi_per_node
        """
        return self._requires

    @property
    def name(self) -> str:
        """Returns the name under which this resource is stored"""
        return self._name

    @property
    def flag(self):
        """Returns the flag set for the jobscript"""
        return self._flag

    @property
    def min(self):
        """Minimal numeric value"""
        return self._min

    @property
    def max(self):
        """Maximal numeric value"""
        return self._max

    @property
    def default(self):
        """Returns the default, if available"""
        return self._value.default

    @default.setter
    def default(self, default):
        self._value.default = default

    @property
    def value(self):
        """Returns the set value, otherwise returns the default"""
        if self.default is not None and self._value is None:
            val = self.default
        else:
            val = self._value

        try:
            val = val.value
        except AttributeError:
            pass

        try:
            val / 1
            isnumeric = True

            if val < 1:
                val = 1
        except TypeError:
            isnumeric = False

        if self.format == "float":
            return float(val)
        if self.format == "time":
            return format_time(val)

        if isnumeric:
            return math.ceil(val)
        return val

    @value.setter
    def value(self, value):
        """Sets the value"""
        try:
            value / 1
            isnumeric = True
        except TypeError:
            isnumeric = False

        if isnumeric:
            if self.min is not None and value < self.min:
                raise ValueError(
                    f"{value} for {self.flag} is less than minimum value {self.min}"
                )
            if self.max is not None and value > self.max:
                raise ValueError(
                    f"{value} for {self.flag} is more than maximum value {self.max}"
                )

        if isinstance(self._value, DynamicValue):
            try:
                self._value.value = value
                return
            except RuntimeError:
                warnings.warn(f"Dynamic chain broken when assigning {self.name}")

        self._value = DynamicValue(value)

    @property
    def resource_line(self) -> str:
        """
        Shortcut to output a suitable resource request line

        Returns:
            str: resource request line
        """
        pragma = f"{self.pragma} " if self.pragma is not None else ""
        tag = self.tag if self.tag is not None else "--"
        separator = self.separator if self.separator is not None else "="

        return f"{pragma}{tag}{self.flag}{separator}{self.value}"


class runargs(dict):
    """
    Class to contain the dataset run_args in a way that won't break any loops
    over the resources

    Args:
        args (dict):
            Dataset run_args
    """

    _accesserror = (
        "\nParser is attempting to access the flag of the run_args, you "
        "should add an `if {option}: ...` catch to your parser."
        "\nRemember that placeholders without an argument are 'falsy', "
        "see the docs for more info. https://l_sim.gitlab.io/remotemanager"
        "/remotemanager.connection.computers.options.html"
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __bool__(self):
        return False

    @property
    def value(self):
        """
        Prevents an AttributeError when a parser attempts to access the value.

        Returns:
            (dict): internal dict
        """
        return self.__dict__

    @property
    def flag(self):
        """
        Parsers should not access the flag method of the run_args, doing so likely
        means that a loop has iterated over this object and is attempting to insert
        it into a jobscript.

        Converts an AttributeError to one more tailored to the situation.

        Returns:
            RuntimeError
        """
        raise RuntimeError(runargs._accesserror)


class Resources:
    """
    Container class to store Resource objects for use by a parser
    """

    __slots__ = ["_names", "_resources", "_run_args", "pragma"]

    def __init__(self, resources, pragma, tag, separator, run_args):
        self._names = []
        self._resources = resources
        self._run_args = run_args

        self.pragma = pragma

        for resource in resources:
            self._names.append(resource.name)
            # add pragma to Resource for resource_line property
            resource.pragma = pragma
            if resource.tag is None:
                resource.tag = tag
            if resource.separator is None:
                resource.separator = separator

    def __iter__(self):
        return iter(self._resources)

    def __getitem__(self, item: str) -> Union[Resource, dict]:
        """
        Need to enable Resources["mpi"], for example

        Args:
            item:
                name of resource to get

        Returns:
            Resource
        """
        if item == "run_args":
            return self.run_args
        try:
            return self._resources[self._names.index(item)]
        except ValueError:
            raise ValueError(f"{self} has no resource {item}")

    def get(self, name: str, default: any = "_unspecified"):
        """Allows resource.get(name)"""
        if default == "_unspecified":
            return getattr(self, name)
        return getattr(self, name, default)

    @property
    def run_args(self) -> dict:
        """Returns the stored run_args"""
        return self._run_args
